#import <appkit/appkit.h>
#import <objc/List.h>
#import "AShape.h"
#import "Ellipse.h"
#import "DrawView.h"
#import "TextShape.h"

#define RECTANGLE 0
#define ELLIPSE 1
#define TEXT	2
#define EDIT	3

int	SIMPLEDRAW= YES;

@implementation DrawView

+newFrame:(NXRect const *)r inSV:aScrollView
{
    self = [super newFrame:r];
    fill = NO;
    currentShade = 0.0;
    shapeList = [List new];
    selectionList = [List new];
    scrollview = aScrollView;
    drawType = RECTANGLE;
    return self;
}

-(BOOL)acceptsFirstResponder { return YES;}


-setShadeFromSender:sender
{
    [self setCurrentShade:[sender floatValue]];
    return self;
}
 
-setCurrentShade:(float)s
{
    currentShade = s;
    return self;
}

-setFillFromSender:sender
{
    [self setFill:[sender selectedTag]];
    return self;
}

-setFill:(int)i;
{
    fill = i;
    return self;
}

-setDrawType:sender
{
    drawType = [sender selectedTag];
    return self;
}

-windowWillReturnFieldEditor:win toObject:anObject
{
    if(fe==nil){
	fe = [Text new];
	[fe setCharFilter:NXFieldFilter];
    }
    [self changeOwner:anObject];
    return fe;
}

-changeOwner:newOwner
{
    if(((currentOwner= [fe delegate])!=nil)&&
	([currentOwner respondsTo:@selector(textDidEnd:endChar:)]))
	    [currentOwner textDidEnd:fe endChar:'0'];
    [fe setDelegate:newOwner];
    currentOwner = newOwner;
    return self;
}

-changeFont:sender
{
    int i =[selectionList count];
    id sO;
    
    while(i){
    	i--;
     	if([(sO=[selectionList objectAt:i]) respondsTo:@selector(changeFont:)])
	    [sO changeFont:sender];
    }
    [self display];
    return self;
}

-addSelection:anObject
{
    [selectionList addObject:anObject];
    [self lockFocus];
	[anObject setSelection:YES];
    [self unlockFocus];
    [window flushWindow];
}

-emptySelection:anObject
{
    [self emptySelection:anObject andRemove:NO];
    return self;
}

-emptySelection:anObject andRemove:(BOOL)remove
{
    id sO;
    NXRect r={0.,0.,0.,0.};
    int s = [selectionList count];
    int firstTime = YES;
    
    while(s){
	s--;
	sO = [selectionList objectAt:s];
	if(firstTime){
	    r = *[sO bbox];
	    firstTime = NO;
	}
	else
	    NXUnionRect([sO bbox],&r);
	[sO setSelection:NO];
	if(remove)
	    [[shapeList removeObject:sO] free];
    }
    [selectionList empty];
    NXInsetRect(&r,-10.,-10);
    [self display:&r:1:NO];
    return self;
}

-removeSelection:anObject
{
    NXRect r;
    [[selectionList removeObject:anObject] setSelection:NO];
    r = *[anObject bbox];
    NXInsetRect(&r,-10.,-10.);
    [self display:&r :1 :NO];
    return self;
}

-pt:(NXPoint *)pt inList:(id)whichList
{
    id sel=nil;
     
    int s = [whichList count];
    while(s){
	s--;
	if([[whichList objectAt:s]containsPt:pt]){
	    sel= [whichList objectAt:s];
	    break;
	}
    }
    return sel;
}

-selectObjectsWithinRect:(NXRect *)r
{
    int s = [shapeList count];
    id co;
    
    [window disableFlushWindow];
    while(s){
	s--;
	co = [shapeList objectAt:s];
	if(NXIntersectsRect([co bbox],r))
	    [self addSelection:co];
    }
    [window reenableFlushWindow];
    return self;
} 
    	    
-makeNewShape:(NXRect *)bbox;
{
    id	newShape;
    
    switch(drawType){
    case RECTANGLE:
	newShape=[AShape newShapeWithBBox:bbox];
	break;
    case ELLIPSE:
	newShape=[Ellipse newShapeWithBBox:bbox];
	break;
    case TEXT:
	newShape = [TextShape newShapeWithBBox:bbox];
	break;
    case EDIT:
	// NXAlert("Edit Not Implemented Yet","OK",NULL,NULL);
	newShape=[AShape newShapeWithBBox:bbox];
	break;
    }
    if([newShape needsView])
	[newShape setView:self];
    if(drawType !=EDIT){
	[newShape setFill:fill];
	[newShape setShade:currentShade];
	[newShape setSelection:YES];
    }
    else{
	[newShape setFill:0];
	[newShape setShade:.666];
    }
    return newShape;
}

-addShape:newObject
{
    if (newObject !=nil){
	[shapeList addObject:newObject];
    }
    return newObject;
}

-dragOutShape:(id)aShape startingRect:(NXRect *)r2
{
    NXEvent *e;
    NXPoint	sPt,aPt;
    
    e=[NXApp getNextEvent:NX_MOUSEUPMASK|NX_MOUSEDRAGGEDMASK];
    while (e->type==NX_MOUSEDRAGGED){
	aPt = e->location;
	[self convertPoint:&aPt fromView:nil];
	r2->size.width = aPt.x - r2->origin.x;
	r2->size.height = aPt.y - r2->origin.y;
	[aShape setBbox:r2];
	PSnewinstance();   
	[superview autoscroll:e];
	PSsetinstance(YES);
	if(r2->size.height && r2->size.width)
	    [aShape drawObject];
	PSsetinstance(NO);
	e=[NXApp getNextEvent:NX_MOUSEUPMASK|NX_MOUSEDRAGGEDMASK];
    }
    PSnewinstance();
    return self;
}

-editMouseAction:(NXPoint *)Pt
{
    NXEvent *e;
    NXPoint sPt=*Pt;
    id aShape;
    NXRect r2;
    
    if(aShape=[self pt:&sPt inList:selectionList])
    	[self moveSelection];    
    else {
	if(aShape =[self pt:&sPt inList:shapeList]){
	    [self emptySelection:self];
	    [self addSelection:aShape];
	    [aShape edit];
	}
	else {
	    [self emptySelection:self];
	    NXSetRect(&r2,sPt.x,sPt.y,0.0,0.0);
	    aShape = [self makeNewShape:&r2];
	    [self dragOutShape:aShape  startingRect:&r2];
	    massageRect(&r2);
	    [self selectObjectsWithinRect:&r2];
	}
    }
    return self;
}

-mouseDown:(NXEvent *)e
{
    int	looping= YES;
    int	oldMask;
    NXRect	r2;
    NXPoint	sPt,aPt;
    id		aShape;
    
    oldMask = [window addToEventMask:NX_MOUSEDRAGGEDMASK];
    [self lockFocus];
    sPt = e->location;
    [self convertPoint:&sPt fromView:nil];
    if(drawType ==EDIT)
	[self editMouseAction:&sPt];
    else {
	[self emptySelection:self];
	NXSetRect(&r2,sPt.x,sPt.y,0.0,0.0);
	aShape = [self makeNewShape:&r2];
	[self dragOutShape:aShape startingRect:&r2];
	[self addShape:aShape];
	[aShape drawObject];
    }
    [self unlockFocus];
    [window flushWindow];
    [window setEventMask:oldMask];
    if(drawType !=EDIT){
	[self addSelection:aShape];
	[aShape edit];
    }
    return(self);
}
   
-moveSelection
{
    return self;
}
   
-erase:sender
{
    [self lockFocus];
	NXEraseRect(&bounds);
    [self unlockFocus];
    [window flushWindow];
    return self;
}

-clear:sender
{
    [selectionList empty];
    [shapeList freeObjects];
    [self erase:self];
    return self;
}

-reDraw:sender
{
    [self cut:sender];
    return self;
}

-setSimpleDraw:(int)YorN
{
    SIMPLEDRAW = YorN;
    return self;
}

-drawSelf:(NXRect *)r :(int)c;
{
   id cShape;
   NXRect tR;
   int cs,count;
   
   count = [shapeList count];
   if([selectionList count])
       [selectionList makeObjectsPerform:@selector(drawPrep)];
   if(SIMPLEDRAW){
       NXEraseRect(r); 
       for(cs=0;cs<count;cs++)
	   [[shapeList objectAt:cs] drawObject];
   }    
   else {
       NXEraseRect(r);
       NXRectClip(r); 
       for(cs=0;cs<count;cs++){
	   cShape = [shapeList objectAt:cs];
	   if(NXIntersectsRect(r,[cShape bbox]))
	       [cShape drawObject];
	}
    }
}

-cut:sender
{
    const char *const types[1]={"DrawPrivate"};
    id pb = [NXApp pasteboard];
    char * buf;
    int len,c;
        
    // declare types
    [pb declareTypes:types num:1 owner:self];
    
    // writeout selectionList to a typed stream
    buf = NXWriteRootObjectToBuffer(selectionList,&len);
    
    // write from stream into pasteboard
    [pb writeType:types[0] data:buf length:len ];
    
    // free typed stream
    NXFreeObjectBuffer(buf,len);
    
    //empty selectionList and remove objects from shapeList
    [self emptySelection:self andRemove:YES];
        
    return self;
}

-paste:sender
{
    char **type;
    id pb = [NXApp pasteboard];
    char *buf;
    int len,c;
    id tlist,tO;
    
    //disable flush until all objects have been read in
    [window disableFlushWindow];
    [self lockFocus];
    
    // empty selection list
    [self emptySelection:sender];
    
    // check if type is in pasteboard
    for(type = [pb types];*type;type++){
	if(!strcmp(*type,"DrawPrivate"))
	    break;
    }
    if(*type){
	// read from the pasteboard
	[pb readType:*type data:&buf length:&len];
	
	// Use typed stream to read from pasteboard
	tlist = NXReadObjectFromBuffer(buf,len);
    }
    c = [tlist count];
    while(c){
	c--;
	tO = [tlist objectAt:c];
	[self addShape:tO];
	if([tO needsView])
	    [tO setView:self];
	[tO drawShape];
	[self addSelection:tO];
    }
    [tlist empty];
    [self unlockFocus]; 
    [window reenableFlushWindow];
    [window flushWindow];
    return self;
}

    
    
    	            
